from anc.data.anc_processor import Processor
from anc.data.anc_composer import SeqSplitConfig
from anc.data.processors.tokenizer import HFPretrainedTokenizer
import logging
import os
import sys
import torch
import json
from typing import List, Tuple


class FakeOpt:
    def __init__(self, hf_model_name, prompt_style, add_bos):
        self.hf_model_name = hf_model_name
        self.prompt_style = prompt_style
        self.prefix_template = ""
        self.delimiter = ""
        self.assistant_delimiter = None
        self.overall_prefix = None
        self.add_bos = add_bos
        self.do_capitalize = False


def setup_separate_prompt(opt):
    if opt.prompt_style == 'sunhao_old':
        opt.prefix_template = '### {role}:'
        opt.delimiter = '\n\n'
    # elif opt.prompt_style == 'chatml':
    #     opt.prefix_template = '<|im_start|> {role}\n'
    #     opt.delimiter = '<|im_end|>\n'
    elif opt.prompt_style == 'chatml2':
        opt.prefix_template = '<|im_start|>{role}\n'
        opt.delimiter = '<|im_end|>\n'
    elif opt.prompt_style == 'chatml_non_special': # used when <|im_end|> is not a special token
        opt.prefix_template = '<|im_start|> {role}\n'
        opt.delimiter = ' <|im_end|>\n'
    elif opt.prompt_style == 'llama3':
        opt.prefix_template = '<|start_header_id|>{role}<|end_header_id|>\n\n'
        opt.delimiter = '<|eot_id|>'
    elif opt.prompt_style == 'ds_v2':
        opt.prefix_template = '{role}: '
        opt.assistant_delimiter = '<｜end▁of▁sentence｜>'
        opt.delimiter = '\n\n'
        opt.do_capitalize = True
    elif opt.prompt_style == 'qwen':
        opt.prefix_template = '<|im_start|>{role}\n'
        opt.delimiter = '<|im_end|>\n'
    else:
        # Use chatml2 by default
        opt.prefix_template = '<|im_start|>{role}\n'
        opt.delimiter = '<|im_end|>\n'

    opt.delimiter = opt.delimiter.replace('\\n', '\n')
    logging.info(f'Using dialog template:\n{opt.prefix_template}content{opt.delimiter}')


class PretrainProcessor(Processor):
    """Processor class for pre-training language models.
    
    This class handles tokenization, sequence preparation, and batch creation for
    language model pre-training tasks.
    
    Attributes:
        tokenizer: HuggingFace tokenizer instance
        max_seq_len: Maximum allowed sequence length
        batch_size: Number of sequences per batch
        micro_batch_size: Number of sequences per micro-batch
        pad_token_id: Token ID used for padding
        special_tokens: List of special tokens to be added to tokenizer
        special_token_ids: List of token IDs corresponding to special tokens,
            since we may update the tokenizer config file directly to use reserved
            tokens as special tokens, we need to tell processor here to properly 
            handle them
        pad_to_max_seq_len: Whether to pad sequences to max_seq_len
        seq_split_config: Configuration for sequence splitting
    """
    def __init__(
        self,
        hf_model_name,
        max_seq_len,
        add_bos,
        batch_size,
        micro_batch_size,
        ceil_to_nearest=False,
        pad_label_id=None,
        cu_seqlen_use_old_pack=False,
        special_tokens=None,
        special_token_ids=None,
        pad_to_max_seq_len=False,
        seq_split_config=None,
        remove_special_tokens=True,
    ):
        self.opt = FakeOpt(hf_model_name, "", add_bos)
        self.tokenizer = HFPretrainedTokenizer(self.opt)
        self.max_seq_len = max_seq_len
        self.batch_size = batch_size
        self.micro_batch_size = micro_batch_size
        assert self.micro_batch_size == 1
        self.micro_batch = []
        self.pad_seq_length_to_mult = 16
        self.ceil_to_nearest = ceil_to_nearest
        self.pad_label_id = pad_label_id if pad_label_id is not None else self.tokenizer.end_token_id
        self.pad_token_id = self.tokenizer.end_token_id
        self.cu_seqlen_use_old_pack = cu_seqlen_use_old_pack
        self.pad_to_max_seq_len = pad_to_max_seq_len
        self.remove_special_tokens = remove_special_tokens
        if special_token_ids is None:
            special_token_ids = []
        if special_tokens is not None:
            if isinstance(special_tokens, str):
                special_tokens = [special_tokens]
            self.special_tokens = special_tokens
            self.tokenizer.add_special_tokens(special_tokens)
            special_token_ids += [self.tokenizer.txt2vec(i)[0] for i in self.special_tokens]
        else:
            self.special_tokens = []
        visited = set()
        remove_redundant_special_token_ids = []
        for i in special_token_ids:
            i = int(i)
            if i in visited:
                continue
            visited.add(i)
            remove_redundant_special_token_ids.append(i)
        self.special_token_ids = remove_redundant_special_token_ids
        added_tokens_decoder = self.tokenizer.added_tokens_decoder()
        for i in self.special_token_ids:
            assert i in added_tokens_decoder, f"{i} not in {added_tokens_decoder}"
        added_tokens = [added_tokens_decoder[i].content for i in self.special_token_ids]
        logging.warn(f"Added tokens are {added_tokens} with corresponding ids {self.special_token_ids}")
        if seq_split_config is None:
            seq_split_config = SeqSplitConfig()
        self.seq_split_config = seq_split_config
        self.all_masked_sample_cnt = 0
        self.sample_cnt = 0
        data_to_skip_path = os.getenv("ANC_DATA_TO_SKIP_PATH", None)
        # data_to_skip_path is a file that contains the data to skip, each line contains two columns: filepath and row_idx
        if data_to_skip_path is not None and os.path.exists(data_to_skip_path):
            data_to_skip = [line.strip().split() for line in open(data_to_skip_path, 'r')]
            self.data_to_skip = {(row[0], int(row[1])) for row in data_to_skip}
            logging.info(f"will skip {len(self.data_to_skip)} samples listed in {data_to_skip_path}")
        else:
            self.data_to_skip = set()

        try:
            self.too_long_sample_threshold = int(os.getenv("ANC_TOO_LONG_THRESHOLD"))
        except:
            self.too_long_sample_threshold = None

    def transform(self, item, is_last_sample=False):
        """Transform a text item into tokenized sequences with loss masks.
        
        Args:
            item: Dictionary containing 'content' key with text to process
            is_last_sample: Boolean indicating if this is the last sample
            
        Yields:
            Dictionary containing tokenized sequence and loss mask
        """
        if 'filepath' in item and 'row_idx' in item and (item['filepath'], item['row_idx']) in self.data_to_skip:
            logging.warning(f"skipping sample {item['filepath']}:{item['row_idx']} because it is in the skip list")
            return None
        tokens = self.tokenizer.txt2vec(item['content'])
        if self.too_long_sample_threshold is not None and len(tokens) > self.too_long_sample_threshold:
            logging.warning(f"Too long sample threshold is set, and sample length {len(tokens)} longer than threshold {self.too_long_sample_threshold}. Will drop")
            return None
        if len(tokens) > self.max_seq_len and not self.seq_split_config.allow_split:
            logging.warning(f"pretrain sample encounter with token length {len(tokens)} larger than {self.max_seq_len}, will ignore and continue")
            return None
        loss_mask = [1] * len(tokens)
        # set mask between two special tokens <MASK:Z0k7RxyNaZ> and </MASK:Z0k7RxyNaZ> to zero
        # may need more complex logic when we have more special tokens, or mask logic becomes more complex
        set_to_zero = False
        successive_zero_count = 0
        special_token_idxs = []
        for i in range(len(tokens)):
            if tokens[i] in self.special_token_ids:
                special_token_idxs.append(i)
                set_to_zero = not set_to_zero
                loss_mask[i] = 0
                successive_zero_count += 1
                if successive_zero_count >= self.max_seq_len:
                    # The context is too long, ignore this sample
                    logging.warn(f"The context (tokens between masks) is too long: {successive_zero_count}, will ignore current sample")
                    return None
            elif set_to_zero:
                loss_mask[i] = 0
                successive_zero_count += 1
                if successive_zero_count >= self.max_seq_len:
                    # The context is too long, ignore this sample
                    logging.warn(f"The context (tokens between masks) is too long: {successive_zero_count}, will ignore current sample")
                    return None
            else:
                successive_zero_count = 0
        # check if the first special token is the mask end token
        if special_token_idxs and self.tokenizer.vec2txt(tokens[special_token_idxs[0]]).startswith("</"):
            # flip the loss mask
            logging.warn(f"Flipping loss mask for sample from {item.get('filepath', 'unknown')}:{item.get('row_idx', 'unknown')}")
            special_token_idxs = set(special_token_idxs)
            for i in range(len(loss_mask)):
                if i not in special_token_idxs:
                    loss_mask[i] = 1 - loss_mask[i]

        # to remove special tokens from the input_ids
        if self.remove_special_tokens and special_token_idxs:
            special_token_idxs = set(special_token_idxs)
            new_tokens = [tokens[i] for i in range(len(tokens)) if i not in special_token_idxs]
            new_loss_mask = [loss_mask[i] for i in range(len(loss_mask)) if i not in special_token_idxs]
            tokens = new_tokens
            loss_mask = new_loss_mask
            assert len(tokens) == len(loss_mask)

        if sum(loss_mask) == 0:
            self.all_masked_sample_cnt += 1
            if self.sample_cnt >= 1000:
                logging.warn(f"Within last {self.sample_cnt} samples, {self.all_masked_sample_cnt} samples are all masked out.")
                self.all_masked_sample_cnt = 0
                self.sample_cnt = 0
            logging.debug(f"All tokens mask out for text {item['content']}, will ignore and continue")
            return None
        self.try_add_bos_token(tokens, loss_mask)
        self.sample_cnt += 1
        yield from [{"input_ids": tokens, "loss_mask": loss_mask, "text": item['content'], "row_idx": item['row_idx'], "filepath": item['filepath']}]

    # guide the composer to get the correct sequence length of each item
    def get_token_length_fn(self, item):
        """Get the length of a tokenized sequence. 
        Should be override by subclasses if dataset meta changes.
        
        Args:
            item: Dictionary containing 'input_ids' key
        Returns:
            int: Length of the token sequence
        """
        return len(item['input_ids'])

    # guide the composer to split a long sequence
    def split_fn(self, item, split_length):
        """Split a long sequence into two parts at the specified length.
        Should be override by subclasses if dataset meta changes.
        
        Args:
            item: Dictionary containing sequence data
            split_length: Length at which to split the sequence
            
        Returns:
            tuple: (first_part, second_part) of the split sequence
        """
        assert len(item['input_ids']) > split_length
        to_return = {
            'input_ids': item['input_ids'][:split_length],
            'loss_mask': item['loss_mask'][:split_length],
            'text': item['text'],
            'row_idx': item['row_idx'],
            'filepath': item['filepath'],
        }
        to_remain = {
            'input_ids': item['input_ids'][split_length:],
            'loss_mask': item['loss_mask'][split_length:],
            'text': item['text'],
            'row_idx': item['row_idx'],
            'filepath': item['filepath'],
        }
        return to_return, to_remain

    def try_add_bos_token(self, tokens, loss_mask):
        """Add beginning-of-sequence token if needed.
        
        Args:
            tokens: List of token IDs
            loss_mask: List of mask values corresponding to tokens
        """
        if self.opt.add_bos and tokens[0] != self.tokenizer.start_token_id:
            if self.tokenizer.start_token_id is None:
                return
            assert type(self.tokenizer.start_token_id) is int
            tokens.insert(0, self.tokenizer.start_token_id)
            loss_mask.insert(0, 0)  # Don't compute loss for BOS token

    def _compose_one_sequence(self, items):
        """Compose multiple items into a single sequence with necessary metadata.
        
        Args:
            items: Single item or list of items to compose
            
        Returns:
            dict: Dictionary containing:
                - tokens: Combined token IDs
                - labels: Shifted token IDs for language modeling
                - loss_mask: Mask indicating which tokens to compute loss for
                - position_ids: Position embeddings for each token
                - cu_seqlens: Cumulative sequence lengths
                - texts: Original text content if available
        """
        input_ids = []
        loss_mask = []
        labels = []
        position_ids = []
        cu_seqlens = []
        texts = []
        row_idxs = []
        filepaths = []
        valid_token_counts = []
        cur_seqlen = 0

        # Handle single item case
        if not isinstance(items, list):
            items = [items]

        # Combine all items into single sequences
        for item in items:
            input_ids += item['input_ids']
            # Shift labels by 1 for next-token prediction
            labels += item['input_ids'][1:] + [self.pad_label_id]
            loss_mask += item['loss_mask'][1:] + [0]
            cu_seqlens.append(cur_seqlen)
            cur_seqlen += len(item['input_ids'])
            position_ids += list(range(len(item['input_ids'])))
            if 'text' in item:
                texts.append(item['text'])
            if 'row_idx' in item:
                row_idxs.append(item['row_idx'])
            if 'filepath' in item:
                filepaths.append(item['filepath'])
            valid_token_counts.append(sum(item['loss_mask'][1:]))

        cu_seqlens.append(cur_seqlen)
        token_count = len(input_ids)

        # Pad sequence to max_seq_len if required
        if self.pad_to_max_seq_len:
            to_pad = self.max_seq_len - token_count
            if to_pad > 0:
                input_ids += [self.pad_token_id] * to_pad
                labels += [self.pad_label_id] * to_pad
                loss_mask += [0] * to_pad
                position_ids += [0] * to_pad
                token_count = self.max_seq_len

        # Convert lists to tensors
        if loss_mask[0] == 0:
            logging.warn(f"Encounter one micro batch with loss mask starting with 0")
        res = {}
        res['tokens'] = torch.LongTensor(input_ids)
        res['labels'] = torch.LongTensor(labels)
        res['loss_mask'] = torch.LongTensor(loss_mask)
        res['position_ids'] = torch.LongTensor(position_ids)
        res['token_count'] = token_count
        res['cu_seqlens'] = cu_seqlens
        if texts:
            res['texts'] = texts
        if row_idxs:
            res['row_idxs'] = row_idxs
        if filepaths:
            res['filepaths'] = filepaths
        res['valid_token_counts'] = valid_token_counts
        return res

    def _collate_item(self, item, max_length, pad_id):
        """Pad a single item to the specified length.
        
        Args:
            item: Tensor to pad
            max_length: Target length after padding
            pad_id: Token ID to use for padding
            
        Returns:
            torch.Tensor: Padded tensor
        """
        item = torch.nn.functional.pad(
            item, 
            (0, max_length - item.shape[0]), 
            value=pad_id
        )
        return item

    def create_batch(self, items):
        """Create a batch from multiple sequences with proper padding.
        
        Handles padding and creation of attention masks and other metadata needed
        for transformer processing.
        
        Args:
            items: List of sequence dictionaries to batch
            
        Returns:
            dict: Batch dictionary containing:
                - tokens: Padded token sequences
                - labels: Padded label sequences
                - loss_mask: Padded loss masks
                - position_ids: Padded position embeddings
                - attention_mask: Batch attention mask
                - cu_seqlens: Cumulative sequence lengths
                - cu_seqlens_argmin: Indices of sequence starts
                - max_seqlen: Maximum sequence length in batch
        """
        res = {}
        # Find maximum token count in batch for padding
        max_token_count = max([i['token_count'] for i in items])
        if self.ceil_to_nearest:
            max_token_count = self._ceil_to_nearest(
                max_token_count, 
                self.pad_seq_length_to_mult
            )

        # Process each key in the items
        for k in items[0]:
            if k == 'tokens':
                res[k] = torch.stack([
                    self._collate_item(item[k], max_token_count, self.pad_token_id) 
                    for item in items
                ])
            elif k == 'labels':
                res[k] = torch.stack([
                    self._collate_item(item[k], max_token_count, self.pad_label_id) 
                    for item in items
                ])
            elif k in {'loss_mask', 'position_ids'}:
                res[k] = torch.stack([
                    self._collate_item(item[k], max_token_count, 0) 
                    for item in items
                ])
            elif k == 'token_count':
                res[k] = [item['token_count'] for item in items]
            elif k == 'cu_seqlens':
                # Handle cumulative sequence lengths
                for item in items:
                    if item[k][-1] < max_token_count:
                        if self.cu_seqlen_use_old_pack:
                            item[k][-1] = max_token_count
                        else:
                            item[k].append(max_token_count)
                    item[k] = torch.IntTensor(item[k])
                max_seq_count = max(len(item[k]) for item in items)
                res[k] = torch.stack([
                    self._collate_item(item[k], max_seq_count + 1, -1) 
                    for item in items
                ])
            elif k == 'texts':
                res[k] = [item[k] for item in items]
            elif k == 'row_idxs':
                res[k] = [item[k] for item in items]
            elif k == 'filepaths':
                res[k] = [item[k] for item in items]
            elif k == 'valid_token_counts':
                res[k] = [item[k] for item in items]

        # Create attention mask and other metadata
        res['attention_mask'] = torch.LongTensor([1] * len(items))
        res['cu_seqlens_argmin'] = torch.argmin(res['cu_seqlens'], dim=1, keepdim=True)
        seqlens = res['cu_seqlens'][:, 1:] - res['cu_seqlens'][:, :-1]
        max_seqlen, _ = seqlens.max(dim=1, keepdim=True)
        res['max_seqlen'] = max_seqlen
        # create cu_seqlens_unpadded, since we didn't pad any token between sequences 
        # during the sequence composing, we simply clone the cu_seqlens and cu_seqlens_argmin here
        res['cu_seqlens_unpadded'] = res['cu_seqlens'].clone()
        res['cu_seqlens_unpadded_argmin'] = res['cu_seqlens_argmin'].clone()
        return res

    def _ceil_to_nearest(self, n, m):
        """Round up to nearest multiple.
        
        Args:
            n: Number to round up
            m: Multiple to round to
            
        Returns:
            int: n rounded up to nearest multiple of m
        """
        return (n + m - 1) // m * m

    def batch_transform(self, list_of_items, is_last_batch=False):
        """Transform a list of items into batches.
        
        Args:
            list_of_items: List of sequences to batch
            is_last_batch: Whether this is the final batch
            
        Yields:
            dict: Processed batch of sequences
        """
        for items in list_of_items:
            one_sequence = self._compose_one_sequence(items)
            self.micro_batch.append(one_sequence)
            if len(self.micro_batch) == self.batch_size:
                one_batch = self.create_batch(self.micro_batch)
                self.micro_batch = []
                yield one_batch
                
        # Handle any remaining items in the last batch
        if is_last_batch and self.micro_batch:
            one_batch = self.create_batch(self.micro_batch)
            self.micro_batch = []
            yield one_batch

    def __getstate__(self):
        """Get object state for pickling."""
        return self.__dict__
    
    def __setstate__(self, states):
        """Set object state during unpickling."""
        self.__dict__.update(states)


class SFTProcessor(PretrainProcessor):
    """Processor for Supervised Fine-Tuning (SFT) of language models.
    
    Handles processing of conversational data with different roles (user, assistant, etc.)
    for supervised fine-tuning tasks. Inherits from PretrainProcessor.
    """
    def __init__(
        self,
        hf_model_name,
        prompt_style,
        max_seq_len,
        add_bos,
        batch_size,
        micro_batch_size,
        ceil_to_nearest=False,
        pad_label_id=None,
        cu_seqlen_use_old_pack=False,
        pad_to_max_seq_len=True,
        special_tokens=None,
        special_token_ids=None,
    ):
        """Initialize SFT processor with model and processing configurations.
        
        Args:
            hf_model_name: Name of the HuggingFace model to use
            prompt_style: Style template for formatting prompts
            max_seq_len: Maximum sequence length
            add_bos: Whether to add beginning-of-sequence token
            batch_size: Number of sequences per batch
            micro_batch_size: Size of micro-batches
            ceil_to_nearest: Whether to round sequence lengths up
            pad_label_id: ID to use for padding labels
            cu_seqlen_use_old_pack: Whether to use old packing method
        """
        super().__init__(
            hf_model_name,
            max_seq_len,
            add_bos,
            batch_size,
            micro_batch_size,
            ceil_to_nearest,
            pad_label_id,
            cu_seqlen_use_old_pack,
            pad_to_max_seq_len=pad_to_max_seq_len,
            special_tokens=special_tokens,
            special_token_ids=special_token_ids,
        )
        self.opt = FakeOpt(hf_model_name, prompt_style, add_bos)
        setup_separate_prompt(self.opt)

    def transform(self, item, is_last_sample=False):
        """Transform conversation data into model input format.
        
        Args:
            item: Dictionary containing conversation messages
            is_last_sample: Whether this is the last sample
            
        Yields:
            dict: Processed item
        """
        # if this is a pretrain data, delegate to pretrain processor
        if 'text' in item:
            pretrain_data = {"content": item['text']}
            if 'id' in item:
                pretrain_data['id'] = item['id']
            if 'filepath' in item:
                pretrain_data['filepath'] = item['filepath']
            if 'row_idx' in item:
                pretrain_data['row_idx'] = item['row_idx']
            yield from super().transform(pretrain_data, is_last_sample)
            return

        # Validate that all messages have required fields with correct types
        if not all(turn['role'] and turn['content'] and 
                  type(turn['role']) is str and type(turn['content']) is str 
                  for turn in item['messages']):
            logging.warn(f"Found an illegal sample and will skip it...")
            return
            
        # Process conversation context
        context = [(turn['role'].strip(), turn['content'], turn.get('no_loss', False)) for turn in item['messages']]
        last_turn_mask_flag = item.get('last_turn_mask', False) or item.get('loss_on_last_turn', False)
        no_mask_flag = item.get('no_mask', False)
        
        context_vec, loss_mask = self._encode_one(context, last_turn_mask_flag, no_mask_flag)
        if len(context_vec) > self.max_seq_len and not self.seq_split_config.allow_split:
            logging.warn(f"encounter sample with token length {len(context_vec)} larger than {self.max_seq_len}, will ignore and continue")
            return None
        # append one more token at the end of each sequence since nemo would truncate one
        # context_vec.append(self.tokenizer.end_token_id)
        # loss_mask.append(0)
        yield from [{"input_ids": context_vec, "loss_mask": loss_mask, "filepath": item["filepath"], "row_idx": item["row_idx"]}]

    def _encode_one(self, context, last_turn_mask_flag, no_mask_flag):
        """Encode a single conversation into token IDs and loss masks.
        
        Args:
            context: List of (role, content) tuples representing the conversation
            last_turn_mask_flag: Whether to only compute loss on the last turn
            no_mask_flag: Whether to disable masking
            
        Returns:
            tuple: (token_ids, loss_mask)
        """
        context_vec, loss_mask = self._build_chitchat_prompt(
            context,
            self.opt.delimiter,
            loss_on_last_turn=last_turn_mask_flag,
            enable_query_mask=no_mask_flag,
        )
        self.try_add_bos_token(context_vec, loss_mask)
        return context_vec, loss_mask

    def _build_chitchat_prompt(
        self,
        context: List[Tuple[str, str]], 
        dialog_sep: str=None, 
        loss_on_last_turn=False, 
        enable_query_mask=False, 
        add_generation_suffix=False,
        **kwargs
    ):
        """Build a formatted conversation prompt with appropriate masking.
        
        Args:
            context: List of (role, content) tuples for the conversation
            dialog_sep: Separator between dialogue turns
            loss_on_last_turn: Whether to only compute loss on the last turn
            enable_query_mask: Whether to disable masking for queries
            add_generation_suffix: Whether to add assistant prefix for generation
            
        Returns:
            tuple: (token_ids, loss_mask) for the formatted conversation
        """
        assert not (loss_on_last_turn and enable_query_mask)
        if dialog_sep is None:
            dialog_sep: str = self.opt.delimiter

        def _encode_one(prefix: str, content: str, suffix: str, is_output: bool = False):
            """Helper function to encode a single turn of conversation.
            
            Args:
                prefix: Role prefix template
                content: Message content
                suffix: Separator suffix
                is_output: Whether this is an assistant's output
                
            Returns:
                tuple: (role_tokens, content_tokens, suffix_tokens)
            """
            if is_output or enable_query_mask:
                # For assistant outputs or when no_query_mask is True,
                # encode role prefix separately from content
                encoded_role, encoded_content, encoded_suffix = (
                    self.tokenizer.txt2vec(prefix), 
                    self.tokenizer.txt2vec(content + suffix),
                    []
                )
            else:
                # For other roles, encode everything together
                encoded_role, encoded_content, encoded_suffix = (
                    [],
                    self.tokenizer.txt2vec(prefix + content + suffix),
                    []
                )
            return encoded_role, encoded_content, encoded_suffix

        idxs_encoded = []  # List to store all token IDs
        idxs_mask = []    # List to store loss mask values
        prefix_template: str = self.opt.prefix_template
        do_capitalize = self.opt.do_capitalize

        # Process each turn in the conversation
        for i, (role, content, no_loss) in enumerate(context):
            suffix_sep = dialog_sep

            if role == 'assistant':
                if do_capitalize:
                    role = role.capitalize()
                if self.opt.assistant_delimiter is not None:
                    suffix_sep = self.opt.assistant_delimiter
                # Handle assistant's responses
                encoded_role, encoded_content, encoded_suffix = _encode_one(
                    prefix_template.format(role=role), 
                    content, 
                    suffix_sep, 
                    is_output=True
                )
                # Set mask based on loss_on_last_turn flag
                if loss_on_last_turn:
                    to_learn = i == len(context) - 1
                else:
                    to_learn = not no_loss
                content_mask = [int(to_learn)] * len(encoded_content)
            else:
                # Handle other roles (user, system, etc.)
                if role not in ('user', 'human', 'system', 'document', 'memory', 
                              'function', 'ipython', 'tool', 'tool_calls'):
                    logging.warn(f"Found unknown role={role}")
                if do_capitalize:
                    role = role.capitalize()
                encoded_role, encoded_content, encoded_suffix = _encode_one(
                    prefix_template.format(role=role), 
                    content, 
                    suffix_sep
                )
                content_mask = [int(enable_query_mask)] * len(encoded_content)

            # Combine encoded parts and their masks
            idxs_encoded.extend(encoded_role + encoded_content + encoded_suffix)
            idxs_mask.extend([0] * len(encoded_role) + content_mask + [0] * len(encoded_suffix))

        # Add generation prefix if needed
        # incase need to add some token other than bos token
        if self.opt.overall_prefix is not None:
            idxs_encoded = self.tokenizer.txt2vec(self.opt.overall_prefix) + idxs_encoded
            idxs_mask = [0] * (len(idxs_encoded) - len(idxs_mask)) + idxs_mask
        if add_generation_suffix:
            idxs_encoded.extend(self.tokenizer.txt2vec(prefix_template.format(role='assistant')))
            idxs_mask.extend([0] * (len(idxs_encoded) - len(idxs_mask)))

        assert len(idxs_encoded) == len(idxs_mask)
        return idxs_encoded, idxs_mask


class SFTProcessorWithDecode(SFTProcessor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def decode_fn(self, line):
        item = json.loads(line)
        transformed = super().transform(item)
        if transformed is None:
            return []
        res = []
        for item in transformed:
            res.append(item)
        return res

    def transform(self, item, is_last_sample=False):
        yield from [item,]


class PPOProcessor:
    """Processor for handling data in PPO (Proximal Policy Optimization) training.
    
    This processor handles text tokenization and formatting for PPO-based training,
    including sequence length management and batch preparation.
    
    Attributes:
        tokenizer: Tokenizer instance for converting text to token IDs
        add_eod: Whether to append end-of-document token
        max_sample_length: Maximum allowed length for input samples
        collate_fn: Function for combining samples into batches
    """
    def __init__(self, cfg, tokenizer, collate_fn):
        """Initialize PPO processor with configuration and tokenizer.
        
        Args:
            cfg: Configuration object containing model and PPO parameters
            tokenizer: Tokenizer instance for text processing
            collate_fn: Function to collate samples into batches
        """
        self.tokenizer = tokenizer
        self.add_eod = cfg.model.data.get("append_eod", False)
        
        # Calculate maximum sample length based on sequence length and PPO config
        seq_length = cfg.model.data.seq_length
        if "length_params" in cfg.model.ppo:
            # Reserve space for PPO-specific length requirements
            max_sample_length = seq_length - cfg.model.ppo.length_params.max_length
        else:
            # Default to using half the sequence length
            max_sample_length = seq_length // 2
        self.max_sample_length = max_sample_length
        self.collate_fn = collate_fn

    def transform(self, sample, is_last_sample=False):
        """Transform a text sample into model input format.
        
        Converts text to token IDs and applies necessary formatting and length checks.
        
        Args:
            sample: Dictionary containing text data
            is_last_sample: Whether this is the last sample in the dataset
            
        Returns:
            list: List containing processed sample dictionary with:
                - text: Tensor of token IDs
                - length: Length of the sequence
                - loss_multiplier: Flag for loss calculation
                Returns None if sample exceeds maximum length
        """
        text = sample["text"]
        # Convert text to token IDs
        text_ids = self.tokenizer.text_to_ids(text)
        
        # Append end-of-document token if needed
        if len(text_ids) > 0 and self.add_eod:
            text_ids.append(self.tokenizer.eos_id)
            
        # Skip samples that exceed maximum length
        if len(text_ids) > self.max_sample_length:
            return
            
        # Convert to tensor and prepare output format
        sample_tensor = torch.tensor(text_ids, dtype=torch.int64)
        output = {
            "text": sample_tensor,
            "length": sample_tensor.shape[0],
            "loss_multiplier": True,  # Flag for loss calculation
        }
        return [output]
    
    def batch_transform(self, items, is_last_batch=False):
        """Transform a list of items into a batch.
        
        Args:
            items: List of processed samples to batch
            is_last_batch: Whether this is the final batch
            
        Returns:
            list: List containing single batched item created by collate_fn
        """
        return [self.collate_fn(items)]


class DPOProcessor(SFTProcessor):
    """Processor for Direct Preference Optimization (DPO) training.
    
    Handles processing of paired responses with preference rankings for DPO training.
    Inherits from SFTProcessor and adds preference-specific processing capabilities.
    
    Attributes:
        ensure_common_tokens: Whether to mask out common tokens between responses
        default_chosen_reward: Default reward value for chosen responses
        default_rejected_reward: Default reward value for rejected responses
        pad_reward_id: Padding value for reward tensors
    """
    def __init__(
        self,
        hf_model_name,
        prompt_style,
        max_seq_len,
        add_bos,
        batch_size,
        micro_batch_size,
        ceil_to_nearest=False,
        ensure_common_tokens=False,
        default_chosen_reward=1.0,
        default_rejected_reward=0.0,
        pad_label_id=-100,
        pad_reward_id=-1000,
        cu_seqlen_use_old_pack=False,
    ):
        """Initialize DPO processor with model and training configurations."""
        super().__init__(
            hf_model_name,
            prompt_style,
            max_seq_len,
            add_bos,
            batch_size,
            micro_batch_size,
            ceil_to_nearest,
            pad_label_id,
            cu_seqlen_use_old_pack,
        )
        self.ensure_common_tokens = ensure_common_tokens
        self.default_chosen_reward = default_chosen_reward
        self.default_rejected_reward = default_rejected_reward
        self.pad_reward_id = pad_reward_id
    
    def decode_fn(self, line):
        """Decode a JSON line into training examples with paired responses.
        
        Args:
            line: JSON string containing conversation and response data
            
        Returns:
            list: List of dictionaries containing messages and paired responses with ranks
        """
        session = json.loads(line)
        edited, chosen, rejected = session.get('edited', None), session.get('chosen', None), session.get('rejected', None)
        responses = []
        output = []
        
        # Collect valid responses
        if edited and edited.get('content'):
            responses.append(edited['content'])
        if chosen and chosen.get('content'):
            responses.append(chosen['content'])
        if rejected and rejected.get('content'):
            responses.append(rejected['content'])
            
        # Skip if not enough responses for comparison
        if len(responses) < 2:
            return output
        # Skip if last message is from assistant
        if session['messages'][-1]['role'] == 'assistant':
            return output

        # Create pairs of responses for comparison
        for i_yw in range(len(responses) - 1):
            for i_yl in range(i_yw + 1, len(responses)):
                if responses[i_yw] != responses[i_yl]:
                    output.append({
                        'messages': session['messages'],
                        'responses': [responses[i_yw], responses[i_yl]],
                        'ranks': [0, 1],
                    })
        return output

    def mask_common_token(self, yw_text_vec, yw_loss_mask, yl_text_vec, yl_loss_mask):
        """Mask out common tokens at the start and end of responses.
        
        Args:
            yw_text_vec: Token IDs for first response
            yw_loss_mask: Loss mask for first response
            yl_text_vec: Token IDs for second response
            yl_loss_mask: Loss mask for second response
        """
        # Mask common tokens from the start
        for i in range(min(len(yw_text_vec), len(yl_text_vec))):
            if yw_loss_mask[i] == yl_loss_mask[i] == 0:
                continue
            if yw_text_vec[i] == yl_text_vec[i]:
                yw_loss_mask[i] = yl_loss_mask[i] = -1
            else:
                break

        # Mask common tokens from the end
        for i in range(1, min(len(yw_text_vec), len(yl_text_vec))):
            if yw_loss_mask[-i] == yl_loss_mask[-i] == 0:
                continue
            if yw_text_vec[-i] == yl_text_vec[-i]:
                yw_loss_mask[-i] = yl_loss_mask[-i] = -1
            else:
                break

    def get_token_length_fn(self, item):
        """Get the length of the tokenized sequence.
        
        Args:
            item: input item
            
        Returns:
            int: Total length of the tokenized sequence
        """
        return len(item['yw_input_ids']) + len(item['yl_input_ids'])

    def transform(self, item, is_last_sample=False):
        """Transform a preference pair into model inputs.
        
        Args:
            item: Dictionary containing messages and paired responses
            is_last_sample: Whether this is the last sample
            
        Yields:
            dict: Processed inputs for both responses with rewards
        """
        context = [(turn['role'], turn['content']) for turn in item['messages']]
        responses = item['responses']
        ranks = item['ranks']
        yw_idx, yl_idx = 0, 1  # winner (yw) and loser (yl) indices
        assert ranks[yw_idx] < ranks[yl_idx]
        
        # Process both responses
        yw, yl = responses[yw_idx], responses[yl_idx]
        yw_context_vec, yw_loss_mask = self._encode_one(context + [('assistant', yw)], True, False)
        yl_context_vec, yl_loss_mask = self._encode_one(context + [('assistant', yl)], True, False)
        
        # Optionally mask common tokens
        if self.ensure_common_tokens:
            self.mask_common_token(yw_context_vec, yw_loss_mask, yl_context_vec, yl_loss_mask)
            
        yield from [{
            "yw_input_ids": yw_context_vec,
            "yw_loss_mask": yw_loss_mask,
            "yl_input_ids": yl_context_vec,
            "yl_loss_mask": yl_loss_mask,
            "chosen_reward": item.get("chosen_reward", self.default_chosen_reward),
            "rejected_reward": item.get("rejected_reward", self.default_rejected_reward)
        }]

    def _compose_one_sequence(self, items):
        """Compose a list of items into a single sequence.
        
        Args:
            items: List of dictionaries containing processed inputs
            
        Returns:
            dict: Composed sequence with rewards
        """
        if not isinstance(items, list):
            items = [items]
        res = super()._compose_one_sequence(items)
        res['rewards'] = torch.FloatTensor([i['reward'] for i in items])
        # loss_mask = res.pop('loss_mask')
        loss_mask = res['loss_mask']
        # nemo aligner use label < -1 as the loss mask
        # so we create the masked label inadvance here
        res['labels'][loss_mask==0] = self.pad_label_id
        return res

    def create_batch(self, items):
        """Create a batch from a list of items.
        
        Args:
            items: List of dictionaries containing processed inputs
            
        Returns:
            dict: Batched inputs ready for DPO training
        """
        res = super().create_batch(items)
        max_num_sequences = max(len(item['rewards']) for item in items)
        res['rewards'] = torch.stack([self._collate_item(item['rewards'], max_length=max_num_sequences, pad_id=self.pad_reward_id) for item in items])
        # nemo dpo needs input_ids instead of tokens
        res['input_ids'] = res.pop('tokens')
        return res

    def batch_transform(self, list_of_items, is_last_batch=False):
        """Transform a list of items into batches for DPO training.
        
        Args:
            list_of_items: List of preference pairs to process
            is_last_batch: Whether this is the final batch
            
        Yields:
            dict: Batched inputs ready for DPO training
        """
        for items in list_of_items:
            # Recreate items with proper structure for both chosen and rejected responses
            recreated_items = []
            if not isinstance(items, list):
                items = [items]
            for item in items:
                # Add chosen response
                recreated_items.append({
                    'input_ids': item['yw_input_ids'],
                    'loss_mask': item['yw_loss_mask'],
                    'reward': item['chosen_reward'],
                })
                # Add rejected response
                recreated_items.append({
                    'input_ids': item['yl_input_ids'],
                    'loss_mask': item['yl_loss_mask'],
                    'reward': item['rejected_reward'],
                })
                
            # Create and accumulate batches
            one_sequence = self._compose_one_sequence(recreated_items)
            self.micro_batch.append(one_sequence)
            if len(self.micro_batch) == self.batch_size:
                one_batch = self.create_batch(self.micro_batch)
                self.micro_batch = []
                yield one_batch
                
        # Handle any remaining items in the last batch
        if is_last_batch and self.micro_batch:
            one_batch = self.create_batch(self.micro_batch)
            self.micro_batch = []
            yield one_batch



from transformers import AutoTokenizer
from typing import List
import logging

class HFPretrainedTokenizer():
    def __init__(self, opt):
        self.opt = opt
        logging.info(f"Loading vocab from huggingface {opt.hf_model_name}")
        # if getattr(self.opt, 'model_type') == 'llama':
        #     from transformers.models.llama.tokenization_llama import LlamaTokenizer
        #     self.hf_tokenizer = LlamaTokenizer.from_pretrained(
        #         opt.hf_model_name, 
        #         trust_remote_code=True,
        #         bos_token='<s>',
        #         eos_token='</s>',
        #         unk_token='<unk>',
        #     )
        # else:
        self.hf_tokenizer = AutoTokenizer.from_pretrained(opt.hf_model_name, trust_remote_code=True)

        # Add special tokens
        self.override_special_tokens()
        self.vocab_size = len(self.hf_tokenizer.get_vocab())
        logging.info(f'Special tokens: {self.hf_tokenizer.special_tokens_map},\nwhere {list(zip(self.hf_tokenizer.all_special_tokens, self.hf_tokenizer.all_special_ids))}')
        
    @property
    def gmask_token(self):
        return self.hf_tokenizer.gmask_token
    
    @property
    def gmask_token_id(self):
        return self.hf_tokenizer.gmask_token_id
    
    @property
    def null_token(self):
        if self.hf_tokenizer.pad_token != None:
            return self.hf_tokenizer.pad_token
        elif self.hf_tokenizer.unk_token != None:
            return self.hf_tokenizer.unk_token
        else:
            raise ValueError('Missing pad token')
    
    @property
    def end_token(self):
        return self.hf_tokenizer.eos_token
    
    @property
    def unk_token(self):
        return self.hf_tokenizer.unk_token
    
    @property
    def start_token(self):
        return self.hf_tokenizer.bos_token

    @property
    def cls_token(self):
        return self.hf_tokenizer.cls_token
    
    @property
    def null_token_id(self):
        if self.hf_tokenizer.pad_token_id != None:
            return self.hf_tokenizer.pad_token_id
        elif self.hf_tokenizer.unk_token_id != None:
            return self.hf_tokenizer.unk_token_id
        else:
            raise ValueError('Missing pad token')
    
    @property
    def end_token_id(self):
        return self.hf_tokenizer.eos_token_id
    
    @property
    def unk_token_id(self):
        return self.hf_tokenizer.unk_token_id
    
    @property
    def start_token_id(self):
        return self.hf_tokenizer.bos_token_id

    @property
    def cls_token_id(self):
        return self.hf_tokenizer.cls_token_id
    
    
    def override_special_tokens(self):
        pass
        
    def txt2vec(self, text: str, add_special=False):
        # return [token for token in self.hf_tokenizer.encode(text, add_special_tokens=False) if token != self.unk_token_id]
        # text = self.detokenizer.detokenize(text.split())
        return self.hf_tokenizer.encode(text, add_special_tokens=add_special)

    def vec2txt(self, vector: List[int], skip_special=False):
        text = self.hf_tokenizer.decode(vector, skip_special_tokens=skip_special)
        return text

    def add_special_tokens(self, special_tokens):
        self.hf_tokenizer.add_tokens(special_tokens)

    def added_tokens_decoder(self):
        return self.hf_tokenizer.added_tokens_decoder

    def __len__(self):
        return len(self.hf_tokenizer)


import re
def check_first_sentence(message: str) -> bool:

    if 'npc_speak' not in message:
        return False

    message = message.strip()
    if '\n' in message or '\\n' in message:
        return True
    message = re.sub(r'\[.*?\]', '', message)
    message = message.strip()
    fields = message.split(' ')
    if len(fields) >= 10 and any(punct in ' '.join(fields[-1:]) for punct in ['.', '?', '!', '…']):
        return True

    return False








from torch.utils.data.sampler import *
from torch.utils.data.distributed import DistributedSampler
import random
import torch
import time
import math
import copy
import bisect
from .utils import line_print


class CustomizedRange:
    def __init__(self, start, end, seed=0):
        self.start = start
        self.end = end
        self.seed = seed
        self.random = random.Random(seed)

    def __len__(self):
        return self.end - self.start

    def clone(self, clone_times=1):
        return CustomizedRange(self.start, self.end, self.seed + clone_times)

    def to_list(self, shuffle=False):
        lst = list(range(self.start, self.end))
        if shuffle:
            self.random.shuffle(lst)
        return lst

    def split_to_lists(self, chunk_sizes, shuffle=False):
        lst = self.to_list(shuffle)
        chunks = []
        assert sum(chunk_sizes) == self.end - self.start
        accu_length = 0
        for chunk_size in chunk_sizes:
            chunks.append(lst[accu_length: accu_length + chunk_size])
            accu_length += chunk_size
        return chunks

    def slice(self, start, length):
        assert start + self.start + length <= self.end
        return CustomizedRange(start + self.start, start + self.start + length, self.seed)


class RangesHolder:
    def __init__(self, list_of_ranges):
        self.list_of_ranges = list_of_ranges
        self.accu_lengths = self.create_accu_length(list_of_ranges)
        self.length = sum(len(r) for r in list_of_ranges)
        self.expanded = {}
        self.i = 0
    
    def __len__(self):
        return self.length

    def create_accu_length(self, list_of_ranges):
        accu_length = 0
        accu_lengths = []
        for rng in list_of_ranges:
            accu_lengths.append(accu_length)
            accu_length += len(rng)
        accu_lengths.append(accu_length)
        return accu_lengths

    def _get_start_idx(self, start):
        idx = bisect.bisect_right(self.accu_lengths, start) - 1
        assert 0 <= idx < len(self.accu_lengths)
        return idx

    def get_start_and_end_ranges(self, start, length):
        start_range_id, start_range_offset = 0, 0
        n = len(self.list_of_ranges)
        end = start + length
        end_range_id = n - 1
        end_range_offset = len(self.list_of_ranges[end_range_id])
        if self.accu_lengths[self.i] <= start < self.accu_lengths[self.i + 1]:
            start_range_id = self.i
        else:
            start_range_id = self._get_start_idx(start)
        start_range_offset = start - self.accu_lengths[start_range_id]
        accu_length = self.accu_lengths[start_range_id]
        for i in range(start_range_id, n):
            accu_length += len(self.list_of_ranges[i])
            if end < accu_length:
                end_range_id = i
                end_range_offset = end - (accu_length - len(self.list_of_ranges[i]))
                break
        self.i = end_range_id
        return start_range_id, start_range_offset, end_range_id, end_range_offset

    def slice(self, start, length):
        start_range_id, start_range_offset, end_range_id, end_range_offset = self.get_start_and_end_ranges(start, length)
        res = []
        for i in range(start_range_id, end_range_id + 1):
            if i == start_range_id:
                if start_range_id == end_range_id:
                    res.append(self.list_of_ranges[i].slice(start_range_offset, end_range_offset - start_range_offset))
                else:
                    res.append(self.list_of_ranges[i].slice(start_range_offset, len(self.list_of_ranges[i]) - start_range_offset))
            elif i == end_range_id:
                res.append(self.list_of_ranges[i].slice(0, end_range_offset))
            else:
                res.append(self.list_of_ranges[i])
        target = RangesHolder(res)
        assert target.length == length
        return target
    
    def combine(self, other):
        for i in other.list_of_ranges:
            self.list_of_ranges.append(i)
        self.length = sum(len(r) for r in self.list_of_ranges)

    def to_list(self, shuffle=False):
        res = []
        for rng in self.list_of_ranges:
            res += rng.to_list(shuffle)
        return res

    def expand(self, range_idx, shuffle=False):
        if range_idx in self.expanded:
            return self.expanded[range_idx]
        else:
            self.expanded[range_idx] = self.list_of_ranges[range_idx].to_list(shuffle)
            return self.expanded[range_idx]

    def get_value_from(self, start, end, shuffle=False):
        start_range_id, start_range_offset, end_range_id, end_range_offset = self.get_start_and_end_ranges(start, end - start)
        res = []
        for range_idx in range(start_range_id, end_range_id + 1):
            indices = self.expand(range_idx, shuffle)
            if range_idx == start_range_id:
                if range_idx == end_range_id:
                    assert start_range_offset + end - start <= len(self.list_of_ranges[range_idx])
                    res = indices[start_range_offset: start_range_offset + end - start]
                else:
                    res += indices[start_range_offset:]
            elif range_idx == end_range_id:
                res += indices[:end_range_offset]
            else:
                res += indices
        return res

    def random_reset(self, idx):
        self.expanded = {}
        self.list_of_ranges = self.list_of_ranges[idx:] + self.list_of_ranges[:idx]
        self.accu_lengths = self.create_accu_length(self.list_of_ranges)

    

class AncSampler(DistributedSampler):
    r'''
    A batch sampler controls the data sharding, data shuffling and data loading order with proper indices

    The data sharding is chunk based. For example, to shard a dataset with 10 elements into 2 splits, 
    the result data index would be [[0,1,2,3,4],[5,6,7,8,9]] instead of [[0,2,4,6,8],[1,3,5,7,9]]

    Args:
        dataset: dataset from which to load the data.
        batch_size (int): how many samples per batch to load.
        world (int, optional): data parallel world size (default: ``1``).
        rank (int, optional): data parallel rank of current process (default: ``0``).
        num_workers (int): how many subprocesses to use for data
            loading. ``0`` means that the data will be loaded in the main process.
        shuffle (bool, optional): set to ``True`` to have the data reshuffled
            at every epoch (default: ``False``).
        drop_last (bool, optional): set to ``True`` to drop the last incomplete batch,
            if the dataset size is not divisible by the batch size. If ``False`` and
            the size of dataset is not divisible by the batch size, then the last batch
            will be smaller. (default: ``False``).
        seed (int, optional): seed for randomness (default: ``0``).
        resume_step (int, optional): the step to resume from, 
            the previous steps will be skipped (default: ``0``).
        repeat (bool, optional): set to ``True`` to repeat the indices when gone through 
            all the data. If ``False`` than StopIteration will be raised when all data 
            is consumed (default: ``False``).
    '''
    def __init__(
        self,
        dataset,
        batch_size,
        world = 1,
        rank = 0,
        num_workers = 1,
        shuffle = False,
        drop_last = False,
        seed = 0,
        resume_step = 0,
        repeat = False,
        global_shuffle = False,
        ratios = [1],
        chunk_granularity = -1,
        mem_save_mode = False,
        ds_id=0,
        is_train=True,
    ):
        if hasattr(dataset, 'get_sub_lengths'):
            self.sub_lengths = dataset.get_sub_lengths()
        else:
            assert isinstance(dataset, list)
            self.sub_lengths = dataset
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.resume_step = resume_step
        self.step = resume_step
        self.seed = seed
        self.last_iter_epoch = -1
        self.indices = None
        self.repeat = repeat
        self.inner_epoch_count = 0
        self.global_shuffle = global_shuffle
        self.num_replicas = world
        self.rank = rank
        self.epoch = 0
        self.drop_last = drop_last
        self.chunk_granularity = chunk_granularity
        self.mem_save_mode = mem_save_mode
        self.ds_id = ds_id
        self.is_train = is_train
        if self.mem_save_mode:
            assert self.chunk_granularity == -1, "chunk_granularity is not supported when mem_save_mode is True"
        if len(ratios) == 1:
            self.ratios = ratios * len(self.sub_lengths)
        else:
            assert len(ratios) == len(self.sub_lengths)
            self.ratios = ratios
        ds_length = 0
        for item, ratio in zip(self.sub_lengths, self.ratios):
            if isinstance(item, list):
                ds_length += math.ceil(sum(item) * ratio)
            else:
                ds_length += math.ceil(item * ratio)
        if self.drop_last and ds_length % self.num_replicas != 0:
            self.num_samples = math.ceil((ds_length - self.num_replicas) / self.num_replicas)
        else:
            self.num_samples = math.ceil(ds_length / self.num_replicas)
        self.total_size = self.num_samples * self.num_replicas
        self.shuffle = shuffle
        if getattr(self, "dataset", None) is not None:
            self.dataset = None
        self.random = random.Random()

    def _get_indices_with_ratio(self, indices, ratio):
        ratio_int = math.floor(ratio)
        ratio_float = ratio - ratio_int
        res = []
        for ratio_cnt in range(ratio_int):
            if isinstance(indices, CustomizedRange):
                tmp_indices = indices.clone(ratio_cnt)
                res.append(tmp_indices)
            else:
                tmp_indices = indices[:]
                if self.shuffle:
                    self.random.shuffle(tmp_indices)
                if self.chunk_granularity > 0:
                    num_chunks = (len(tmp_indices) + self.chunk_granularity - 1) // self.chunk_granularity
                    for i in range(num_chunks):
                        res.append(tmp_indices[i * self.chunk_granularity: (i + 1) * self.chunk_granularity])
                else:
                    res.append(tmp_indices)
        if ratio_float > 0:
            if isinstance(indices, CustomizedRange):
                tmp_indices = indices.clone(ratio_int)
                length = math.ceil(ratio_float * len(tmp_indices))
                start = 0 if not self.shuffle else self.random.randint(0, len(tmp_indices) - length)
                res.append(tmp_indices.slice(start, length))
            else:
                tmp_indices = indices[:]
                if self.shuffle:
                    self.random.shuffle(tmp_indices)
                tmp_indices = tmp_indices[:math.ceil(ratio_float * len(tmp_indices))]
                if self.chunk_granularity > 0:
                    num_chunks = (len(tmp_indices) + self.chunk_granularity - 1) // self.chunk_granularity
                    for i in range(num_chunks):
                        res.append(tmp_indices[i * self.chunk_granularity: (i + 1) * self.chunk_granularity])
                else:
                    res.append(tmp_indices)
        if self.shuffle:
            self.random.shuffle(res)
        return res

    def _get_indices(self, sub_lengths, offset=0, return_sub_indices=False, ratio=[1]):
        if self.mem_save_mode:
            assert return_sub_indices is True
        indices_list = []
        self.random.seed(self.seed)
        accumulate_length = offset
        if len(ratio) == 1:
            ratio = ratio * len(sub_lengths)
        else:
            assert len(ratio) == len(sub_lengths)
        for i, item in enumerate(sub_lengths):
            if not self.mem_save_mode:
                sub_indices = list(range(accumulate_length, accumulate_length + item))
            else:
                sub_indices = CustomizedRange(accumulate_length, accumulate_length + item)
            sub_indices = self._get_indices_with_ratio(sub_indices, ratio[i])
            indices_list += sub_indices
            accumulate_length += item

        if self.shuffle:
            self.random.shuffle(indices_list)
        if return_sub_indices:
            return indices_list
        indices = []
        for item in indices_list:
            indices += item
        return indices

    def create_chunk_indices_from_bz_and_worker(self, indices):
        worker_info = torch.utils.data.get_worker_info()
        #if os.getenv("WID_FOR_DEBUG") is not None:
        #    actual_wid = int(os.getenv("WID_FOR_DEBUG"))
        #else:
        #    actual_wid = worker_info.id if worker_info is not None else 0
        actual_wid = worker_info.id if worker_info is not None else 0
        # calculate num batches per worker
        num_batches = (self.num_samples + self.batch_size - 1) // self.batch_size
        last_batch_size = self.num_samples - (num_batches - 1) * self.batch_size
        num_batches_per_worker = num_batches // self.num_workers
        remain = num_batches - self.num_workers * num_batches_per_worker
        num_batches_per_each_worker = [num_batches_per_worker + 1] * remain + [num_batches_per_worker] * (self.num_workers - remain)
        assert sum(num_batches_per_each_worker) == num_batches
        last_batch_worker_id = self.num_workers - 1 if remain == 0 else remain - 1

        num_samples_per_worker = []
        for i in range(self.num_workers):
            if i == last_batch_worker_id:
                current_num_samples = (num_batches_per_each_worker[i] - 1) * self.batch_size + last_batch_size
            else:
                current_num_samples = num_batches_per_each_worker[i] * self.batch_size
            num_samples_per_worker.append(current_num_samples)
        assert sum(num_samples_per_worker) == self.num_samples
        start_idx_per_worker = [0] * self.num_workers
        for i in range(1, self.num_workers):
            start_idx_per_worker[i] = start_idx_per_worker[i - 1] + num_samples_per_worker[i - 1]

        if self.repeat and self.shuffle:
            self.random.seed(self.seed)
        inner_epoch_count = 1
        while True:
            self.inner_epoch_count += 1
            # current_idx_per_worker = [i for i in start_idx_per_worker]
            # for i in range(num_batches):
            #     wid = i % self.num_workers
            #     start_idx = current_idx_per_worker[wid]
            #     end_idx = start_idx + self.batch_size
            #     is_last_batch = (end_idx - start_idx_per_worker[wid]) >= num_samples_per_worker[wid] and not self.repeat
            #     end_idx = min(end_idx, start_idx_per_worker[wid] + num_samples_per_worker[wid])
            #     if self.mem_save_mode:
            #         yield indices.get_value_from(start_idx, end_idx, self.shuffle), is_last_batch
            #     else:
            #         yield indices[start_idx: end_idx], is_last_batch
            #     current_idx_per_worker[wid] = end_idx

            _start_idx = start_idx_per_worker[actual_wid]
            for i in range(num_batches_per_each_worker[actual_wid]):
                _end_idx = _start_idx + self.batch_size
                is_last_batch = (_end_idx - start_idx_per_worker[actual_wid]) >= num_samples_per_worker[actual_wid] and not self.repeat
                _end_idx = min(_end_idx, start_idx_per_worker[actual_wid] + num_samples_per_worker[actual_wid])
                if self.mem_save_mode:
                    yield indices.get_value_from(_start_idx, _end_idx, self.shuffle), is_last_batch
                else:
                    yield indices[_start_idx: _end_idx], is_last_batch
                _start_idx = _end_idx

            if self.is_train:
                line_print(f"dataset {self.ds_id} on rank {self.rank} wid {actual_wid} get into {inner_epoch_count + 1} epoch")
            inner_epoch_count += 1
            if not self.repeat:
                break
            if self.global_shuffle:
                assert self.mem_save_mode is False
                self.random.shuffle(indices)
            elif self.shuffle:
                if self.mem_save_mode:
                    shift_offset = self.random.randint(0, len(indices.list_of_ranges))
                    indices.random_reset(shift_offset)
                else:
                    shift_offset = self.random.randint(0, len(indices))
                    indices = indices[shift_offset:] + indices[:shift_offset]
    
    def __iter__(self):
        indices = []
        if self.epoch == self.last_iter_epoch:
            indices = self.indices
        else:
            self.last_iter_epoch = self.epoch
            if isinstance(self.sub_lengths[0], list):
                sub_indices_list = []
                accumulate_length = 0
                for i, item in enumerate(self.sub_lengths):
                    sub_indices = self._get_indices(item, accumulate_length, True, [self.ratios[i]])
                    accumulate_length += sum(item)
                    sub_indices_list += sub_indices
                if self.shuffle:
                    self.random.shuffle(sub_indices_list)
                if not self.mem_save_mode:
                    for item in sub_indices_list:
                        indices += item
                else:
                    indices = RangesHolder(sub_indices_list)
            else:
                indices = self._get_indices(self.sub_lengths, ratio=self.ratios)
                if self.global_shuffle:
                    self.random.shuffle(indices)
            
            if not self.drop_last:
                # add extra samples to make it evenly divisible
                padding_size = self.total_size - len(indices)
                if padding_size <= len(indices):
                    if self.mem_save_mode:
                        indices.combine(indices.slice(0, padding_size))
                    else:
                        indices += indices[:padding_size]
                else:
                    assert self.mem_save_mode is False, "mem_save_mode is True, but padding_size is greater than indices length, which is not supported now"
                    indices += (indices * math.ceil(padding_size / len(indices)))[
                        :padding_size
                    ]
            else:
                # remove tail of data to make it evenly divisible.
                if self.mem_save_mode:
                    indices = indices.slice(0, self.total_size)
                else:
                    indices = indices[: self.total_size]
            assert len(indices) == self.total_size

            # we do the chunk sharding here
            if self.mem_save_mode:
                indices = indices.slice(self.rank * self.num_samples, self.num_samples)
                # indices = indices.to_list(self.shuffle)
            else:
                indices = indices[self.rank * self.num_samples : (self.rank + 1) * self.num_samples]
            self.indices = indices
        assert len(indices) == self.num_samples
        # further split the per rank indices into num_workers splits
        indices_generator = self.create_chunk_indices_from_bz_and_worker(indices)

        # skip the indices that already been consumed
        for i in range(self.resume_step):
            _ = next(indices_generator)

        # # yield empty indices to let workers follow the original order
        # # otherwise, the first generated indices would be consumed by worker 0
        # # while they should be consumed by cur_wid
        # cur_wid = self.resume_step % self.num_workers
        # for i in range(cur_wid):
        #     yield [], False

        while True:
            try:
                self.step += 1
                yield next(indices_generator)
            except StopIteration:
                break

    def set_step(self, step):
        self.resume_step = step
        self.step = step
    
    def __len__(self):
        # this is a batch sampler, return the number of batches
        if not self.repeat:
            if self.drop_last:
                return self.num_samples // self.batch_size
            else:
                return (self.num_samples + self.batch_size - 1) // self.batch_size
        else:
            return 1000000000000

    def _set_ckpt(self, state):
        self.__dict__.update(state)

    def _get_ckpt(self):
        indices = self.__dict__.pop('indices', None)
        ckpt_state = copy.deepcopy(self.__dict__)
        ckpt_state['resume_step'] = self.step
        ckpt_state['last_iter_epoch'] = -1
        self.__dict__['indices'] = indices
        return ckpt_state


class AncMultiSourceSampler():
    def __init__(
        self,
        dataset,
        ratios,
        batch_size,
        world = 1,
        rank = 0,
        num_workers = 1,
        shuffle = False,
        drop_last = False,
        seed = 0,
        resume_step = 0,
        repeat = False,
        global_shuffle = False,
        chunk_granularity = -1,
        mem_save_mode = False,
        is_train = True,
        initial_step_info=None,
    ):
        self.ratios = ratios
        self.seed = seed + rank
        assert repeat is True
        assert 0.9999 <= sum(ratios) <= 1.0001
        self.samplers = [AncSampler(
            dataset.get_sub_lengths(i),
            batch_size,
            world=world,
            rank=rank,
            num_workers=num_workers,
            shuffle=shuffle,
            drop_last=drop_last,
            seed=seed,
            resume_step=resume_step,
            repeat=repeat,
            global_shuffle=global_shuffle,
            chunk_granularity=chunk_granularity,
            mem_save_mode=mem_save_mode,
            ds_id=i,
            is_train=is_train,
        ) for i in range(len(dataset))]
        self.random = random.Random(self.seed)
        self.shuffle = shuffle
        self.resume_from_ckpt = False
        self.is_train = is_train
        self.initial_step_info = initial_step_info

    def __iter__(self):
        worker_info = torch.utils.data.get_worker_info()
        wid = worker_info.id if worker_info is not None else 0
        num_workers = worker_info.num_workers if worker_info is not None else 1
        if self.resume_from_ckpt:
            self.initial_step_info = None  # use step info from ckpt
        if self.is_train and self.initial_step_info is not None:
            '''self.initial_step_info looks like:
                {
                    0: [20, 22, 24, 21],
                    4: [10, 12, 14, 11],
                    "seed": 0,
                }
            '''
            for i, sampler in enumerate(self.samplers):
                if i in self.initial_step_info:
                    assert len(self.initial_step_info[i]) == num_workers
                    sampler.set_step(self.initial_step_info[i][wid])
                    sampler.seed = self.initial_step_info["seed"]
                    print(f"Setting {i}th sampler step to {self.initial_step_info[i][wid]} in worker id {wid}, setting seed to {self.initial_step_info['seed']}")
        iters = [iter(sampler) for sampler in self.samplers]
        iter_list = list(range(len(iters)))
        if self.resume_from_ckpt:
            self.resume_from_ckpt = False
        elif True or not self.shuffle:
            # if shuffle is False, we want every epoch sampler output the same indices
            self.random = random.Random(self.seed + (wid + 123) * 100)
        while True:
            cur_iter_idx = self.random.choices(iter_list, weights=self.ratios)[0]
            cur_iter = iters[cur_iter_idx]
            indices, is_last_batch = next(cur_iter)
            yield cur_iter_idx, indices, is_last_batch

    def __len__(self):
        return 1000000000000

    def _get_ckpt(self):
        state = {}
        samplers_state = [sampler._get_ckpt() for sampler in self.samplers]
        state['samplers'] = samplers_state
        state['random_state'] = self.random.getstate()
        return state

    def _set_ckpt(self, state):
        assert len(state['samplers']) == len(self.samplers)
        for i, sampler in enumerate(self.samplers):
            sampler._set_ckpt(state['samplers'][i])
        self.random.setstate(state['random_state'])
        self.resume_from_ckpt = True



import torch
import multiprocessing as mp
import os
import glob
import argparse
import json

def get_step_info_from_ckpt(item):
    ckpt_path, target_paths = item
    ckpt = torch.load(ckpt_path, weights_only=False)
    ds_states = ckpt['ds_states']
    paths = ckpt['paths']
    res = {}
    seed_info = set()
    for i, path in enumerate(paths):
        ds_name = path[0].rsplit("/", 2)[0]
        if ds_name in target_paths:
            res[ds_name] = [ds_state['sampler']['samplers'][i]['step'] for ds_state in ds_states]
            for ds_state in ds_states:
                seed_info.add(ds_state['sampler']['samplers'][i]['seed'])
    assert len(seed_info) == 1, "all seeds must be the same"
    res['seed'] = list(seed_info)[0]
    return ckpt_path, res


def get_all_step_info(ckpt_dir, target_paths):
    ckpt_paths = glob.glob(os.path.join(ckpt_dir, '*.pt'))
    inputs = [(ckpt_path, target_paths) for ckpt_path in ckpt_paths]
    with mp.Pool(16) as pool:
        res = pool.map(get_step_info_from_ckpt, inputs)
    res_dict = {}
    for ckpt_path, step_info in res:
        rank_id = int(os.path.basename(ckpt_path).split('.')[0][2:])
        res_dict[rank_id] = step_info
    return res_dict


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--ckpt_dir', type=str, required=True)
    parser.add_argument('--target_names', type=str, required=True)
    parser.add_argument('--output_path', type=str, required=True)
    parser.add_argument('--target_folder', type=str, required=True)
    args = parser.parse_args()
    target_paths = [os.path.join(args.target_folder, name) for name in args.target_names.split(',')]
    res = get_all_step_info(args.ckpt_dir, target_paths)
    with open(args.output_path, 'w') as f:
        json.dump(res, f)



import abc
import subprocess
import time
import requests
from typing import List, Optional
import os

import logging
logger = logging.getLogger("anc")

class BaseServer(abc.ABC):
    """Abstract base class for all server implementations."""
    
    @abc.abstractmethod
    def start(self):
        """Start the server."""
        pass
    
    def check_server_ready(self, host: str, port: int, health_endpoint: str = "/health", 
                          max_retries: int = 300, retry_interval: int = 20) -> bool:
        """
        Check if the server is ready by polling the health endpoint.
        
        Args:
            host: Server host
            port: Server port
            health_endpoint: Health check endpoint
            max_retries: Maximum number of retries
            retry_interval: Time between retries in seconds
            
        Returns:
            bool: True if server is ready, False otherwise
        """
        for retry in range(max_retries):
            try:
                print(f"Checking server health status at {host}:{port}{health_endpoint} (attempt {retry+1}/{max_retries})...")
                # Try to connect to server health check endpoint
                response = requests.get(f"http://{host}:{port}{health_endpoint}")
                if response.status_code == 200:
                    print(f"Server at {host}:{port} is ready")
                    return True
            except requests.exceptions.ConnectionError:
                pass
            
            time.sleep(retry_interval)
        
        print(f"Timed out waiting for server at {host}:{port} to be ready after {max_retries} attempts")
        return False


class ServerFactory:
    """Factory class for creating different types of model serving servers."""
    
    @staticmethod
    def create_server(server_type, **kwargs):
        """
        Create and return a server instance based on the server type.
        
        Args:
            server_type: Type of server to create (vllm, tensorrt-llm, sglang, etc.)
            **kwargs: Arguments specific to the server type
            
        Returns:
            A server instance
        """
        if server_type == "vllm" or server_type == "openai-chat":
            return VLLMServer(**kwargs)
        elif server_type == "tensorrt-llm":
            return TensorRTLLMServer(**kwargs)
        elif server_type == "sglang":
            return SGLangServer(**kwargs)
        elif server_type == "scalellm":
            return ScaleLLMServer(**kwargs)
        else:
            raise ValueError(f"Unsupported server type: {server_type}")


class VLLMServer(BaseServer):
    """VLLM server implementation."""
    
    def __init__(self, model=None, host=None, port=None, 
                **kwargs):
        self.model = model
        self.host = host
        self.port = port
        self.kwargs = kwargs
        self.process = None  # Add this line to store process handle
        
    def start(self, dry_run, server_log_dir, start_otlp_server=False):
        """Start the VLLM server."""
        try:
            # Construct command to start vLLM server
            cmd = [
                "python3", "-m", "vllm.entrypoints.openai.api_server",
                "--model", self.model,
                "--host", self.host,
                "--port", str(self.port),
                "--tensor-parallel-size", str(self.kwargs.get('tensor_parallel_size', 4)),
                # "--enable-chunked-prefill", str(self.kwargs.get('enable_chunked_prefill', "True")),
                "--gpu-memory-utilization", str(self.kwargs.get('gpu_memory_utilization', 0.95)),
                "--max-model-len", str(self.kwargs.get('max-model-len', 8000)),
                # "--multi-step-stream-outputs", str(self.kwargs.get('multi_step_stream_outputs', "True")),
            ]
            if self.kwargs.get('use_v2_block_manager'):
                cmd.append("--use-v2-block-manager")
            if self.kwargs.get('max-num-seqs'):
                cmd.append("--max-num-seqs")
                cmd.append(self.kwargs.get('max-num-seqs'))
            
            if self.kwargs.get('enable_prefix_caching') and self.kwargs.get('enable_prefix_caching') == "True":
                cmd.append("--enable-prefix-caching")
            
            if self.kwargs.get('speculative-model'):
                cmd.append("--speculative-model")
                cmd.append(self.kwargs.get('speculative-model'))
                cmd.append("--speculative-draft-tensor-parallel-size")
                cmd.append(self.kwargs.get('speculative-draft-tensor-parallel-size'))
                cmd.append("--num-speculative-tokens")
                cmd.append(self.kwargs.get('num-speculative-tokens'))
                cmd.append("--speculative-disable-by-batch-size")
                cmd.append(self.kwargs.get('speculative_disable_by_batchsize'))
            
            if start_otlp_server:
                os.environ["OTEL_SERVICE_NAME"] = "vllm-server"
                os.environ["OTEL_EXPORTER_OTLP_TRACES_INSECURE"] = "true"
                cmd.append("--collect-detailed-traces")
                cmd.append("all")
                cmd.append("--otlp-traces-endpoint")
                cmd.append("grpc://localhost:4317")
            if self.kwargs.get('extra_vllm_args'):
                # Split the extra args by space and add each part individually
                extra_args = self.kwargs.get('extra_vllm_args').split()
                cmd.extend(extra_args)
            # Check if port is already in use and kill the process
            self.release_port
            kill_gpu_processes()
                
            logger.info(f"Starting vLLM server with command: {' '.join(cmd)}")
            logger.debug(f"env: {os.environ}")
            if dry_run:
                return
            # Start server as a background process
            # If start_server_only is specified, run server in foreground
            log_file_path = f"{server_log_dir}/vllm_server_{int(time.time())}.log"
            with open(log_file_path, "w") as log_file:
                print(f"\n{'-'*100}")
                print(f"\n{'-'*100}")
                print(f"Server logs will be saved to {log_file_path}")
                print(f"{'-'*100}\n")
                print(f"{'-'*100}\n")
                if self.kwargs.get('start_server_only'):
                    # Run server in foreground with output to both log file and screen
                    self.process = subprocess.Popen(
                        cmd,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                        text=True,
                        bufsize=1
                    )
                    
                    # Print output to both log file and screen
                    for line in self.process.stdout:
                        print(line, end='')
                        log_file.write(line)
                        log_file.flush()
                    
                    self.process.wait()
                    return self.process.returncode == 0
                else:
                    # Run server in background
                    self.process = subprocess.Popen(
                        cmd,
                        stdout=log_file,
                        stderr=subprocess.STDOUT,
                        text=True
                )
                
                # Check if server is ready
                server_ready = self.check_server_ready(self.host, self.port)
                if not server_ready:
                    raise Exception("vLLM server failed to start")
        except Exception as e:
            print(f"Error starting vLLM server: {e}")
            raise e

    def release_port(self):
        try:
            import socket
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            result = sock.connect_ex((self.host, self.port))
            sock.close()
                
            if result == 0:  # Port is in use
                print(f"Port {self.port} is already in use. Attempting to free it...")
                # Propagate the exception to the caller
                raise Exception(f"Port {self.port} is already in use. not able to release it. try to kill the process manually with lsof -ti :{self.port} | xargs kill -9")

        except ImportError:
            print("Warning: psutil not installed. Cannot automatically free port.")
            print("If the server fails to start, manually kill the process using the port.")
        except Exception as e:
            print(f"Error releasing port: {e}")
            raise e

    def stop(self):
        """Stop the VLLM server gracefully."""
        if self.process:
            print("Stopping vLLM server...")
            self.process.terminate()  # Send SIGTERM
            try:
                self.process.wait(timeout=30)  # Wait up to 30 seconds
            except subprocess.TimeoutExpired:
                print("Server didn't terminate gracefully, forcing shutdown...")
                self.process.kill()  # Send SIGKILL
                kill_gpu_processes()  # Kill any remaining GPU processes
            self.process = None


class TensorRTLLMServer(BaseServer):
    """TensorRT-LLM server implementation."""
    
    def __init__(self, host=None, port=None, **kwargs):
        self.host = host
        self.port = port
        self.kwargs = kwargs
        
    def start(self):
        """Start the TensorRT-LLM server."""
        print("Starting TensorRT-LLM server (not yet implemented)")
        # When implemented, should check if server is ready
        # self.check_server_ready(self.host, self.port)
        return False


class SGLangServer(BaseServer):
    """SGLang server implementation."""
    
    def __init__(self, host=None, port=None, **kwargs):
        self.host = host
        self.port = port
        self.kwargs = kwargs
        
    def start(self):
        """Start the SGLang server."""
        print("Starting SGLang server (not yet implemented)")
        # When implemented, should check if server is ready
        # self.check_server_ready(self.host, self.port)
        return False


class ScaleLLMServer(BaseServer):
    """ScaleLLM server implementation."""
    
    def __init__(self, host=None, port=None, **kwargs):
        self.host = host
        self.port = port
        self.kwargs = kwargs
        
    def start(self):
        """Start the ScaleLLM server."""
        print("Starting ScaleLLM server (not yet implemented)")
        # When implemented, should check if server is ready
        # self.check_server_ready(self.host, self.port)
        return False

def kill_gpu_processes():
    """Kill all Python processes using GPU."""
    try:
        # Get all GPU processes
        nvidia_smi = "nvidia-smi"
        process = subprocess.run([nvidia_smi, "--query-compute-apps=pid", "--format=csv,noheader"], 
                               capture_output=True, text=True)
        
        # Extract PIDs
        pids = process.stdout.strip().split('\n')
        
        # Kill each process
        for pid in pids:
            if pid:  # Skip empty lines
                try:
                    subprocess.run(['kill', '-9', pid.strip()])
                    print(f"Killed GPU process {pid}")
                except subprocess.CalledProcessError:
                    print(f"Failed to kill process {pid}")
                    
    except subprocess.CalledProcessError as e:
        print(f"Error getting GPU processes: {e}")
